/*
								+---------------------------------+
								|                                 |
								|    ***   Bitmap font 2   ***    |
								|                                 |
								|  Copyright   -tHE SWINe- 2007  |
								|                                 |
								|          BitmapFont.h           |
								|                                 |
								+---------------------------------+
*/

/*
 *	passed code revision at 2007-06-04
 *
 *	changed bitmap font zoom style so when zoomed to 1 / pixel size, user gets perfectly crisp
 *	letters (in case font rendering functions are passed pixel aligned coordinates)
 *
 *	2007-07-10
 *
 *	added CBitmapFont::SetTextureFilters() to set texture filtering mode in order to improve
 *	rendering of pixel-accurate font
 *	moved function objects sources to BitmapFont.cpp in order to clean-up the header
 *
 *	2007-07-16
 *
 *	updated so now it works with new berLame CGLTexture_2D constructors (texture target
 *	was omitted from constructor parameter list)
 *
 *	added BITMAP_FONT_USE_CLEARTYPE directive for enabling cleartype font support
 *	note (color) cleartype glyphs are converted to greyscale (alpha) but it gives better
 *	results than antialiassed font quality anyway
 *
 *	2007-07-17
 *
 *	added line descent and height font parameters, ie virtual CGLFont::f_LineDescent() const,
 *	virtual CGLFont::f_LineHeight() as well as CGLBitmapFont::f_LineDescent() and
 *	CGLBitmapFont::f_LineHeight()
 *
 */

#ifdef FONT_TEXTURE_CREATE_FONTS

#define BITMAP_FONT_USE_CLEARTYPE
// use cleartype font renderer
// note result is converted to greyscale (alpha), but still it gives better results
// than antialiassed fonts and maybe even better results than cleartype on rotated LCD

#endif

#ifndef __FONTENG_INCLUDED
#define __FONTENG_INCLUDED

class CGLFont {
public:
	/*
	 *	virtual bool CGLFont::Load(CGLState *p_state, const char *p_s_font_name)
	 *		- loads font from source identified as p_s_font_name
	 *		- p_state is OpenGL state manager (new textures are created)
	 *		- returns true on success, false on failure
	 */
	virtual bool Load(CGLState *p_state, const char *p_s_font_name) = 0;

	/*
	 *	virtual void CGLFont::Free()
	 *		- cleanup font, can't be used anymore (should be _ASSERT-ed)
	 */
	virtual void Free() = 0;

	/*
	 *	virtual void CGLFont::Zoom(float f_new_size = 1)
	 *		- scales font to f_new_size (not times f_new_size)
	 */
	virtual void Zoom(float f_new_size = 1) = 0;

	/*
	 *	virtual float CGLFont::f_Zoom() const
	 *		- returns current font scale (default 1)
	 */
	virtual float f_Zoom() const = 0;

	/*
	 *	virtual void CGLFont::Plot_Character(CGLState *p_state,
	 *		Vector2f v_pos, char n_char) const
	 *		- plots a single character n_char to position v_pos; p_state is OpenGL state guard
	 */
	virtual void Plot_Character(CGLState *p_state, Vector2f v_pos, char c_char) const = 0;

	/*
	 *	virtual void CGLFont::Plot_String(CGLState *p_state, Vector2f v_pos,
	 *		const char *p_s_string, int n_length = -1) const
	 *		- plots string of n_length characters from p_s_string (in case n_length is -1,
	 *		  the whole string is used) to position v_pos; p_state is OpenGL state guard
	 *		- note this should be made faster than calling Plot_Character() in loop 
	 */
	virtual void Plot_String(CGLState *p_state, Vector2f v_pos,
		const char *p_s_string, int n_length = -1) const = 0;

	/*
	 *	virtual float CGLFont::f_StringWidth(const char *p_s_string, int n_length = -1) const
	 *		- returns width of n_length characters from p_s_string (in case n_length is -1,
	 *		  the whole string is used)
	 *		- note in case string contains multiple lines of text, the longest line is returned
	 */
	virtual float f_StringWidth(const char *p_s_string, int n_length = -1) const = 0;

	/*
	 *	virtual float CGLFont::f_LineHeight() const
	 *		- returns complete height of line of text including internal / external leading
	 *		- note this should be positive number
	 */
	virtual float f_LineHeight() const = 0;

	/*
	 *	virtual float CGLFont::f_LineDescent() const
	 *		- returns line descension, ie. height of character parts under baseline
	 *		  (characters sucha as gjpqy)
	 *		- note this should be negative number
	 */
	virtual float f_LineDescent() const = 0;

	/*
	 *	virtual float CGLFont::f_InternalLeading() const
	 *		- returns internal leading, ie. height of punctation marks
	 *		  (included in character height)
	 *		- note this should be negative number
	 */
	virtual float f_InternalLeading() const = 0;

	/*
	 *	virtual Vector2f CGLFont::v_CharacterOffset(const char *p_s_string,
	 *		int n_length = -1) const
	 *		- returns offset of n_length-th character from p_s_string (in case n_length is -1,
	 *		  returns offset of the last character)
	 *		- (note offset is vector pointing from first character of the string to the character)
	 */
	virtual Vector2f v_CharacterOffset(const char *p_s_string, int n_length = -1) const = 0;
};

#endif // __FONTENG_INCLUDED

#ifndef __BITMAP_FONT_INCLUDED
#define __BITMAP_FONT_INCLUDED

class CGLBitmapFont : public CGLFont {
protected:
	std::vector<TCharacterInfo> m_char_list;
	std::vector<CGLTexture_2D*> m_texture_list;
	unsigned char m_n_first_char;
	unsigned char m_n_char_num;
	int m_n_space_width;
	int m_n_newline_height;
	int m_n_char_descent, m_n_int_leading;
	float m_f_pixel_x, m_f_pixel_y;
	bool m_b_snap_to_pixels;

	struct TCharLookup {
		GLuint n_display_list; // draw geometry, shift modelview. no glBegin() or glBindTexture()  
		const TCharacterInfo *p_character;
		CGLTexture_2D *p_tex_page;
	};

	TCharLookup *m_p_char_lookup;
	float m_f_zoom;

	class CSetTextureFilters;
	class CDeleteObjects;
	template <class CInnerObject> class CDeref;
	class CFindMin_Count;
	class CFillCharLookup;

public:
	/*
	 *	CGLBitmapFont::CGLBitmapFont()
	 *		- default constructor
	 */
	CGLBitmapFont();

	/*
	 *	CGLBitmapFont::~CGLBitmapFont()
	 *		- default destructor
	 */
	~CGLBitmapFont();

	void EnableSnapToPixels(float f_pixel_x, float f_pixel_y)
	{
		m_f_pixel_x = f_pixel_x;
		m_f_pixel_y = f_pixel_y;
		m_b_snap_to_pixels = true;
	}

	void DisableSnapToPixels()
	{
		m_b_snap_to_pixels = false;
	}

	/*
	 *	virtual bool CGLBitmapFont::Load(CGLState *p_state, const char *p_s_font_name)
	 *		- loads font from p_s_font_name bitmap (or more bitmaps)
	 *		- in case p_s_font_name ends with "_page00.tga", there must me at least
	 *		  one more page with characters (ie. "_page01.tga")
	 *		- p_state is OpenGL state manager (new textures are created)
	 *		- returns true on success, false on failure
	 */
	virtual bool Load(CGLState *p_state, const char *p_s_font_name);

#ifdef FONT_TEXTURE_CREATE_FONTS

	/*
	 *	virtual bool CGLBitmapFont::Create(const char *p_s_font_name, CGLState *p_state)
	 *		- loads font from p_s_font_name bitmap (or more bitmaps)
	 *		- in case p_s_font_name ends with "_page00.tga", there must me at least
	 *		  one more page with characters (ie. "_page01.tga")
	 *		- p_state is OpenGL state manager (new textures are created)
	 *		- returns true on success, false on failure
	 */
	virtual bool Create(CGLState *p_state, const char *p_s_font_face, int n_font_size,
		int n_font_weight = 400, bool b_italic = false, bool b_underline = false,
		int n_char_padding = 4);

#endif //FONT_TEXTURE_CREATE_FONTS

	/*
	 *	virtual void CGLBitmapFont::Free()
	 *		- cleanup font, can't be used anymore
	 */
	virtual void Free();

	/*
	 *	void CGLBitmapFont::SetTextureFilters(CGLState *p_state,
	 *		GLenum n_min_filter = GL_LINEAR_MIPMAP_LINEAR, GLenum n_mag_filter = GL_LINEAR)
	 *		- sets texture filtering mode
	 *		- p_state is OpenGL state guard
	 *		- n_min_filter and n_mag_filter are minification and magnification filters respectively
	 *		- useful for setting GL_NEAREST filtering for pixel-accurate fonts
	 */
	void SetTextureFilters(CGLState *p_state, GLenum n_min_filter = GL_LINEAR_MIPMAP_LINEAR,
		GLenum n_mag_filter = GL_LINEAR);

	/*
	 *	virtual void CGLBitmapFont::Zoom(float f_new_size = 1)
	 *		- zooms font to a given scale
	 */
	virtual void Zoom(float f_new_size = 1);

	/*
	 *	virtual float CGLBitmapFont::f_Zoom() const
	 *		- returns font scale
	 */
	virtual float f_Zoom() const;
	/*
	 *	virtual void CGLBitmapFont::Plot_Character(CGLState *p_state,
	 *		Vector2f v_pos, char n_char) const
	 *		- plots a single character n_char to position v_pos; p_state is OpenGL state guard
	 */
	virtual void Plot_Character(CGLState *p_state, Vector2f v_pos, char n_char) const;

	/*
	 *	virtual void CGLBitmapFont::Plot_String(CGLState *p_state, Vector2f v_pos,
	 *		const char *p_s_string, int n_length = -1) const
	 *		- plots string of n_length characters from p_s_string (in case n_length is -1,
	 *		  the whole string is used) to position v_pos; p_state is OpenGL state guard
	 *		- note this is faster than calling Plot_Character() in loop can be 
	 */
	virtual void Plot_String(CGLState *p_state, Vector2f v_pos,
		const char *p_s_string, int n_length = -1) const;

	/*
	 *	virtual float CGLBitmapFont::f_StringWidth(const char *p_s_string, int n_length = -1) const
	 *		- returns width of n_length characters from p_s_string (in case n_length is -1,
	 *		  the whole string is used)
	 *		- note in case string contains multiple lines of text, the longest line is returned
	 */
	virtual float f_StringWidth(const char *p_s_string, int n_length = -1) const;

	/*
	 *	virtual float CGLBitmapFont::f_LineHeight() const
	 *		- returns complete height of line of text including internal / external leading
	 *		- note this should be positive number
	 */
	virtual float f_LineHeight() const;

	/*
	 *	virtual float CGLBitmapFont::f_LineDescent() const
	 *		- returns line descension, ie. height of character parts under baseline
	 *		  (characters sucha as gjpqy)
	 *		- note this should be negative number
	 */
	virtual float f_LineDescent() const;

	/*
	 *	virtual float CGLBitmapFont::f_InternalLeading() const
	 *		- returns internal leading, ie. height of punctation marks
	 *		  (included in character height)
	 *		- note this should be negative number
	 */
	virtual float f_InternalLeading() const;

	/*
	 *	virtual Vector2f CGLBitmapFont::v_CharacterOffset(const char *p_s_string,
	 *		int n_length = -1) const
	 *		- returns offset of n_length-th character from p_s_string (in case n_length is -1,
	 *		  it's the last character)
	 */
	virtual Vector2f v_CharacterOffset(const char *p_s_string, int n_length = -1) const;

protected:
	int n_Load_Page(const char *p_s_file_name, CGLState *p_state);
};

#endif //__BITMAP_FONT_INCLUDED
