/*
								+---------------------------------+
								|                                 |
								|   ***   Directory lookup  ***   |
								|                                 |
								|  Copyright   -tHE SWINe- 2006  |
								|                                 |
								|              Dir.h              |
								|                                 |
								+---------------------------------+
*/

/*
 *	passed code revision at 2007-07-03
 *
 *	added filename allocation, deallocation and copy functions to TFileInfo
 *	added constructors to TFileInfo::TTime
 *	added TFileInfo::b_valid so it's clear wheter call to CDirectory::Get_FirstFile
 *	or CDirectory::Get_NextFile functions failed or wheter there are no more files in
 *	the directory (as was identified by the same return value before)
 *	merged source code of linux CDirectory and win32 CDirectory together as it shares
 *	substantial sections
 *	created Dir.cpp
 *	added CDirTraversal class
 *
 *	2007-07-09
 *
 *	added TFileInfo copy-constructor TFileInfo(const TFileInfo &r_t_other)
 *	changed CDirTraversal's stack to queue so directory traversal is not reversed
 *
 */

#ifndef __DIR_INCLUDED
#define __DIR_INCLUDED

#ifdef WIN32
#include <windows.h>
#endif

#include <time.h>

/*
 *	struct TFileInfo
 *		- simple file-info struct which keeps memory for it's filename
 */
struct TFileInfo {
protected:
	char *p_s_filename;

public:
	bool b_valid; // are data inside this structure valid?
	bool b_directory; // file / directory flag

	enum {
		flag_Read = 1,
		flag_Write = 2,
		flag_Execute = 4,

		flag_owner_Read = flag_Read,
		flag_owner_Write = flag_Write,
		flag_owner_Execute = flag_Execute,

		flag_group_Read = flag_Read << 3,
		flag_group_Write = flag_Write << 3,
		flag_group_Execute = flag_Execute << 3,

		flag_other_Read = flag_Read << 6,
		flag_other_Write = flag_Write << 6,
		flag_other_Execute = flag_Execute << 6
	};

	int n_flags; // access flags (combination of flag_[owner|group|other]_[Read|Write|Execute])
	unsigned int n_size_lo; // filesize - 32 low bits
	unsigned int n_size_hi; // filesize - 32 high bits

	/*
	 *	struct TFileInfo::TTime
	 *		- very simple time structure (unification for linux / windows)
	 */
	struct TTime {
		short n_month;
		short n_day;
		short n_year;
		short n_hour;
		short n_minute;
		short n_second;

		/*
		 *	TFileInfo::TTime::TTime()
		 *		- default constructor; sets all members to -1
		 */
		inline TTime()
			:n_month(-1), n_day(-1), n_year(-1), n_hour(-1), n_minute(-1), n_second(-1)
		{}

#ifdef WIN32
		/*
		 *	TFileInfo::TTime::TTime(const SYSTEMTIME &r_t_time)
		 *		- (win32) conversion constructor, taking struct SYSTEMTIME as input
		 *		- always succeeds
		 */
		TTime(const SYSTEMTIME &r_t_time);

		/*
		 *	TFileInfo::TTime::TTime(const FILETIME &r_t_time)
		 *		- (win32) conversion constructor, taking struct FILETIME as input
		 *		- note FILETIME needs to be converted to SYSTEMTIME first,
		 *		  in case conversion fails, all members are set to -1
		 */
		TTime(const FILETIME &r_t_time);
#else //WIN32
		/*
		 *	TFileInfo::TTime::TTime(const tm *p_time)
		 *		- (unix) conversion constructor, taking struct tm as input
		 *		- always succeeds
		 */
		TTime(const tm *p_time);
#endif //WIN32

		/*
		 *	bool TFileInfo::TTime::operator ==(const TTime &r_t_time) const
		 *		- equality operator
		 *		- returns true in case this is equal to r_t_time, otherwise false
		 */
		bool operator ==(const TTime &r_t_time) const;
	};

	enum {
		time_Creation = 0,
		time_LastAccess,
		time_LastWrite
	};

	TTime p_time[3]; // file time (use time_[Creation|LastAccess|LastWrite] as index)

	/*
	 *	inline TFileInfo::TFileInfo()
	 *		- default constructor
	 */
	inline TFileInfo()
		:p_s_filename(0)
	{}

	/*
	 *	inline TFileInfo::TFileInfo(const TFileInfo &r_other)
	 *		- copy constructor
	 */
	inline TFileInfo(const TFileInfo &r_other)
		:p_s_filename(0)
	{
		*this = r_other; // copy values
	}

	/*
	 *	inline TFileInfo::~TFileInfo()
	 *		- default destructor
	 */
	inline ~TFileInfo()
	{
		if(p_s_filename)
			delete[] p_s_filename;
	}

	/*
	 *	inline const char *TFileInfo::p_s_Filename() const
	 *		- returns filename as zero-terminated string
	 *		  or 0 in case no filename was specified yet
	 */
	inline const char *p_s_Filename() const
	{
		return p_s_filename;
	}

	/*
	 *	bool TFileInfo::SetFilename(const char *p_s_new_filename)
	 *		- sets filename to p_s_new_filename
	 *		- returns true on success, false on failure
	 *		- note p_s_new_filename can be 0 (sets this filename to 0 as well)
	 */
	bool SetFilename(const char *p_s_new_filename);

	/*
	 *	bool TFileInfo::operator =(const TFileInfo &r_t_file_info)
	 *		- copy-operator
	 *		- returns true on success, false in case there
	 *		  was not enough memory to hold filename string
	 */
	bool operator =(const TFileInfo &r_t_file_info);
};

#ifdef WIN32

//#include <windows.h> // included earlier
#include <string.h>
#include <io.h>

#else //WIN32

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <string.h>

#endif //WIN32

/*
 *	class CDirectory
 *		- simple directory class featuring simple access to it's files (directories)
 */
class CDirectory {
protected:
	char *m_p_s_dir_name;

#ifdef WIN32
	HANDLE m_h_prev_file;
#else //WIN32
    DIR *m_p_dir;
#endif //WIN32

public:
	/*
	 *	CDirectory(const char *p_s_dir_name)
	 *		- default constructor, creates handle to directory with address p_s_dir_name
	 *		  (such as "c:\blah" or "c:\blah\" or ".\blah" or ".\blah\" or "blah" or "blah\")
	 *		- note p_s_dir_name is copied and can be freed right after constructor returns
	 *		- see p_s_Name() to find out wheter there was enough memory for string copy
	 */
	CDirectory(const char *p_s_dir_name);

	/*
	 *	~CDirectory()
	 *		- default destructor
	 */
	~CDirectory();

	/*
	 *	inline const char *p_s_Name() const
	 *		- returns directory name as passed to constructor
	 *		- note it returns 0 in case there was not enough memory for string copy
	 */
	inline const char *p_s_Name() const
	{
		return m_p_s_dir_name;
	}

	/*
	 *	bool Get_FirstFile(TFileInfo &r_t_file)
	 *		- gets first file in the directory, copies it's data to r_t_file
	 *		- in case the directory doesn't have any files, r_t_file.b_valid is set to false
	 *		- returns true on success, false on failure (possibly inaccesible directory)
	 */
	bool Get_FirstFile(TFileInfo &r_t_file);

	/*
	 *	bool Get_NextFile(TFileInfo &r_t_file)
	 *		- gets next file in the directory, copies it's data to r_t_file
	 *		- in case the directory doesn't have any more files, r_t_file.b_valid is set to false
	 *		- returns true on success, false on failure (possibly inaccesible directory)
	 *		- in case Get_FirstFile was not called prior calling Get_NextFile, it's called
	 *		  automatically. in order to get file list for this folder again, Get_FirstFile
	 *		  must be called (r_t_file contents are only written to, not read)
	 */
	bool Get_NextFile(TFileInfo &r_t_file);

protected:
#ifdef WIN32
	bool GetFileInfo(const WIN32_FIND_DATA &r_t_find_data, TFileInfo &r_t_file);
#endif //WIN32
};

/*
 *	class CDirTraversal
 *		- simple non-recursive directory traversal
 */
class CDirTraversal {
public:
	/*
	 *	template <class CFileListener>
	 *	static bool Traverse(const char *p_s_dir_name, CFileListener &r_file_listener)
	 *		- traverse directory p_s_dir_name along with it's subdirectories
	 *		- all files are passed to r_file_listener
	 *		- all files and subdirectories in a directory are passed to r_file_listener before
	 *		  traversing into any of subdirectories
	 */
	template <class CFileListener>
	static bool Traverse(const char *p_s_dir_name, CFileListener &r_file_listener)
	{
		CDirectory *p_dir_info;
		if(!(p_dir_info = new CDirectory(p_s_dir_name)))
			return false;
		// prepare the first directory handle

		std::vector<CDirectory*> m_dir_queue;
		m_dir_queue.reserve(16);
		if(m_dir_queue.capacity() < 1)
			return false;
		m_dir_queue.push_back(p_dir_info);
		// put it into the queue

		while(!m_dir_queue.empty()) {
			CDirectory *p_dir = m_dir_queue.back();
			m_dir_queue.erase(m_dir_queue.end() - 1, m_dir_queue.end());
			// fetch a single record from top of the queue

			unsigned long n_pos = m_dir_queue.size();

			TFileInfo t_file_info;
			if(!p_dir->Get_FirstFile(t_file_info)) {
				std::for_each(m_dir_queue.begin(), m_dir_queue.end(), DeleteDir);
				return false; // t_odo - free the pair data
			}
			while(t_file_info.b_valid) {
				r_file_listener((const TFileInfo&)t_file_info);
				// pass file data to listener

				_ASSERTE(t_file_info.b_valid);
				// must not clear the valid flag
			
				if(t_file_info.b_directory) {
					m_dir_queue.reserve(m_dir_queue.size() + 1);
					CDirectory *p_subdir_info;
					if(m_dir_queue.capacity() < m_dir_queue.size() + 1 ||
					   !(p_subdir_info = new CDirectory(t_file_info.p_s_Filename()))) {
						std::for_each(m_dir_queue.begin(), m_dir_queue.end(), DeleteDir);
						return false;
					}
					// make sure there's enough space in the queue,
					// prepare the subdirectory handle

					m_dir_queue.insert(m_dir_queue.begin() + n_pos, p_subdir_info);
				}
				// add the directory to the list to recurse

				if(!p_dir->Get_NextFile(t_file_info)) {
					std::for_each(m_dir_queue.begin(), m_dir_queue.end(), DeleteDir);
					return false;
				}
				// get next file info
			}
			delete p_dir;
		}

		return true;
	}

protected:
	static inline void DeleteDir(CDirectory *p_dir)
	{
		delete p_dir;
	}
};

#endif //__DIR_INCLUDED
